package de.desy.acop.video.analysis;

import java.util.Arrays;

import org.apache.commons.math.FunctionEvaluationException;
import org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction;
import org.apache.commons.math.analysis.MultivariateMatrixFunction;
import org.apache.commons.math.analysis.MultivariateRealFunction;
import org.apache.commons.math.optimization.GoalType;
import org.apache.commons.math.optimization.OptimizationException;
import org.apache.commons.math.optimization.RealPointValuePair;
import org.apache.commons.math.optimization.VectorialPointValuePair;
import org.apache.commons.math.optimization.direct.MultiDirectional;
import org.apache.commons.math.optimization.general.LevenbergMarquardtOptimizer;
import org.apache.commons.math.util.MathUtils;


/**
 * <code>FittingUtilities</code> provides a set of utilities to fit a gaussian function to a set of data
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
@SuppressWarnings("deprecation")
public class FittingUtilities {

    private static class GaussWithOffset implements DifferentiableMultivariateVectorialFunction {

        private double[] x;

        /**
         * Constructs a new GaussWithOffset, where the values are known for the given horizontal axis.
         * 
         * @param x the horiontal axis values
         */
        public GaussWithOffset(double[] x) {
            this.x = x;
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
         */
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double[][] J = new double[x.length][a.length];
                    for (int i = 0; i < J.length; i++) {
                        J[i][0] = Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2]));
                        J[i][1] = -J[i][0] * a[0] * (x[i] - a[2]) * (x[i] - a[2]);
                        J[i][2] = a[0] * J[i][0] * a[1] * 2 * (x[i] - a[2]);
                        J[i][3] = 1;
                    }
                    return J;
                }
            };
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
         */
        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double[] f = new double[x.length];
            for (int i = 0; i < x.length; i++) {
                f[i] = a[0] * Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2])) + a[3];
            }
            return f;
        }
    }

    private static class GaussWithLinear implements DifferentiableMultivariateVectorialFunction {

        private double[] x;

        /**
         * Constructs a new GaussWithOffset, where the values are known for the given horizontal axis.
         * 
         * @param x the horiontal axis values
         */
        public GaussWithLinear(double[] x) {
            this.x = x;
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
         */
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double[][] J = new double[x.length][a.length];
                    for (int i = 0; i < J.length; i++) {
                        J[i][0] = Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2]));
                        J[i][1] = -J[i][0] * a[0] * (x[i] - a[2]) * (x[i] - a[2]);
                        J[i][2] = a[0] * J[i][0] * a[1] * 2 * (x[i] - a[2]);
                        J[i][3] = 1;
                        J[i][4] = x[i];
                    }
                    return J;
                }
            };
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
         */
        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double[] f = new double[x.length];
            for (int i = 0; i < x.length; i++) {
                f[i] = a[0] * Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2])) + a[3] + a[4] * x[i];
            }
            return f;
        }
    }

    private static class Gauss implements DifferentiableMultivariateVectorialFunction {

        private double[] x;

        /**
         * Constructs a new GaussWithOffset, where the values are known for the given horizontal axis.
         * 
         * @param x the horiontal axis values
         */
        public Gauss(double[] x) {
            this.x = x;
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
         */
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double[][] J = new double[x.length][a.length];
                    for (int i = 0; i < J.length; i++) {
                        J[i][0] = Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2]));
                        J[i][1] = -J[i][0] * a[0] * (x[i] - a[2]) * (x[i] - a[2]);
                        J[i][2] = a[0] * J[i][0] * a[1] * 2 * (x[i] - a[2]);
                    }
                    return J;
                }
            };
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
         */
        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double[] f = new double[x.length];
            for (int i = 0; i < x.length; i++) {
                f[i] = a[0] * Math.exp(-a[1] * (x[i] - a[2]) * (x[i] - a[2]));
            }
            return f;
        }
    }

    private static class InterferometerTrigonometric implements DifferentiableMultivariateVectorialFunction {

        private double[] x;
        private static final double EPSILON = MathUtils.EPSILON;

        public InterferometerTrigonometric(double[] x) {
            this.x = x;
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
         */
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double[][] J = new double[x.length][a.length];
                    double sinc;
                    double cos, sin, xx;
                    for (int i = 0; i < J.length; i++) {
                        J[i][0] = 1;
                        xx = a[2] * (x[i] - a[3]);
                        cos = Math.cos(a[5] * (x[i] - a[6]));
                        sin = Math.sin(a[5] * (x[i] - a[6]));
                        if (Math.abs(xx) < EPSILON) {
                            sinc = 1;
                            J[i][2] = 0; //asin
                            J[i][3] = 0; //psin
                        } else {
                            sinc = Math.sin(xx) / xx;
                            J[i][2] = a[1] * (1 + a[4] * cos) * 2 * sinc * (Math.cos(xx) / xx - sinc / xx)
                                    * (x[i] - a[3]); //asin
                            J[i][3] = -a[1] * (1 + a[4] * cos) * 2 * sinc * (Math.cos(xx) / xx - sinc / xx) * a[2]; //psin
                        }
                        J[i][1] = sinc * sinc * (1 + a[4] * cos); //amp
                        J[i][4] = a[1] * sinc * sinc * cos; //vis
                        J[i][5] = -a[1] * a[4] * sinc * sinc * sin * (x[i] - a[6]); //acos
                        J[i][6] = a[1] * a[4] * sinc * sinc * sin * a[5]; //pcos
                    }
                    return J;
                }
            };
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
         */
        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double[] f = new double[x.length];
            double sinc, xx;
            for (int i = 0; i < x.length; i++) {
                xx = a[2] * (x[i] - a[3]);
                if (Math.abs(xx) < EPSILON) {
                    sinc = 1;
                } else {
                    sinc = Math.sin(xx) / xx;
                }
                f[i] = a[0] + a[1] * sinc * sinc * (1 + a[4] * Math.cos(a[5] * (x[i] - a[6])));
            }
            return f;
        }
    }

    public static class InterferometerCostFunctionTrigonometric implements MultivariateRealFunction {

        private double[] x;
        private double[] y;
        private static final double EPSILON = MathUtils.EPSILON;

        public InterferometerCostFunctionTrigonometric(double[] x, double[] y) {
            this.x = x;
            this.y = y;
        }

        @Override
        public double value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double sinc, xx;
            double sum = 0;
            for (int i = 0; i < x.length; i++) {
                xx = a[2] * (x[i] - a[3]);
                if (Math.abs(xx) < EPSILON) {
                    sinc = 1;
                } else {
                    sinc = Math.sin(xx) / xx;
                }
                xx = a[0] + a[1] * sinc * sinc * (1 + a[4] * Math.cos(a[5] * (x[i] - a[6])));
                sum += (y[i] - xx) * (y[i] - xx);
            }
            return sum;
        }
    }

    public static class InterferometerCostFunctionBessel implements MultivariateRealFunction {

        private double[] x;
        private double[] y;
        private static final double EPSILON = MathUtils.EPSILON;

        public InterferometerCostFunctionBessel(double[] x, double[] y) {
            this.x = x;
            this.y = y;
        }

        @Override
        public double value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double besc, xx;
            double sum = 0;
            for (int i = 0; i < x.length; i++) {
                xx = a[2] * (x[i] - a[3]);
                if (Math.abs(xx) < EPSILON) {
                    besc = 0.5;
                } else {
                    besc = BesselJ.valueAny(1,xx) / xx;
                }
                xx = a[0] + a[1] * besc * besc * (1 + a[4] * Math.cos(a[5] * (x[i] - a[6])));
                sum += (y[i] - xx) * (y[i] - xx);
            }
            return sum;
        }
    }
    
    private static class InterferometerBessel implements DifferentiableMultivariateVectorialFunction {

        private double[] x;
        private static final double EPSILON = MathUtils.EPSILON;

        public InterferometerBessel(double[] x) {
            this.x = x;
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
         */
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double[][] J = new double[x.length][a.length];
                    double cos, sin, xx, besc, bes0, bes2;
                    for (int i = 0; i < J.length; i++) {
                        J[i][0] = 1;
                        xx = a[2] * (x[i] - a[3]);
                        cos = Math.cos(a[5] * (x[i] - a[6]));
                        sin = Math.sin(a[5] * (x[i] - a[6]));
                        if (Math.abs(xx) < EPSILON) {
                            besc = 0.5;
                            J[i][2] = 0; //asin
                            J[i][3] = 0; //psin
                        } else {
                            besc = BesselJ.valueAny(1,xx) / xx;
                            bes0 = BesselJ.valueAny(0,xx);
                            bes2 = BesselJ.valueAny(2,xx);
                            J[i][2] = (a[1]/a[2]) * (1 + a[4] * cos) * besc * (bes0 - bes2 - 2 * besc);
                            J[i][3] = -(a[1]/(x[i] - a[3])) * (1 + a[4] * cos) * besc * (bes0 - bes2 - 2 * besc);
                        }
                        J[i][1] = besc * besc * (1 + a[4] * cos); //amp
                        J[i][4] = a[1] * besc * besc * cos; //vis
                        J[i][5] = -a[1] * a[4] * besc * besc * sin * (x[i] - a[6]); //acos
                        J[i][6] = a[1] * a[4] * besc * besc * sin * a[5]; //pcos
                    }
                    return J;
                }
            };
        }

        /*
         * (non-Javadoc)
         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
         */
        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
            double[] f = new double[x.length];
            double besc, xx;
            for (int i = 0; i < x.length; i++) {
                xx = a[2] * (x[i] - a[3]);
                if (Math.abs(xx) < EPSILON) {
                    besc = 0.5;
                } else {
                    besc = BesselJ.valueAny(1,xx) / xx;
                }
                f[i] = a[0] + a[1] * besc * besc * (1 + a[4] * Math.cos(a[5] * (x[i] - a[6])));
            }
            return f;
        }
    }
    
    private static class Gauss2D implements DifferentiableMultivariateVectorialFunction {

        private final double[] x,y;
        private final int w,h;
        private final double[] f;
        double[][] J;

        public Gauss2D(double[] x, double[] y) {
            this.x = x;
            this.y = y;
            w = x.length;
            h = y.length;
            f = new double[w * h];
            J = new double[w * h][7];
        }

        @Override
        public MultivariateMatrixFunction jacobian() {
            return new MultivariateMatrixFunction() {

                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
                    double d = 0;
                    for (int i = 0; i < w; i++) {
                        for (int j = 0; j < h; j++) {
                            d = Math.exp(-a[1] * (x[i] - a[4]) * (x[i] - a[4])
                                    + 2 * a[2] * (y[j] - a[5]) * (x[i] - a[4]) - a[3] * (y[j] - a[5]) * (y[j] - a[5]));
                            J[j * w + i][0] = d;
                            d *= a[0];
                            J[j * w + i][1] = -d * (x[i] - a[4]) * (x[i] - a[4]);
                            J[j * w + i][2] = d * 2 * (x[i] - a[4]) * (y[j] - a[5]);
                            J[j * w + i][3] = -d * (y[j] - a[5]) * (y[j] - a[5]);
                            J[j * w + i][4] = d * 2 * (a[1] * (x[i] - a[4]) - a[2] * (y[j] - a[5]));
                            J[j * w + i][5] = d * 2 * (a[3] * (y[j] - a[5]) - a[2] * (x[i] - a[4]));
                            J[j * w + i][6] = 1;
                        }
                    }
                    return J;
                }
            };
        }

        @Override
        public double[] value(double[] a) {
            boolean fail = false;
            for (double v : a) {
                if (v < 0) {
                    fail = true;
                    break;
                }
            }
            if (!fail) {
                fail = a[1]*a[3] < a[2] * a[2];
            }
            if (fail) {
                Arrays.fill(f,Double.NaN);
            } else {
                for (int i = 0; i < w; i++) {
                    for (int j = 0; j < h; j++) {
                        f[j * w + i] = a[0] * Math.exp(-a[1] * (x[i] - a[4]) * (x[i] - a[4])
                                + 2 * a[2] * (y[j] - a[5]) * (x[i] - a[4]) - a[3] * (y[j] - a[5]) * (y[j] - a[5])) + a[6];
                    }
                }
            }
            return f;
        }
    }
    
//    private static class SimpleGauss2D implements DifferentiableMultivariateVectorialFunction {
//
//        private final double[] x,y;
//        private final int w,h;
//
//        public SimpleGauss2D(double[] x, double[] y) {
//            this.x = x;
//            this.y = y;
//            w = x.length;
//            h = y.length;
//        }
//
//        /*
//         * (non-Javadoc)
//         * @see org.apache.commons.math.analysis.DifferentiableMultivariateVectorialFunction#jacobian()
//         */
//        public MultivariateMatrixFunction jacobian() {
//            return new MultivariateMatrixFunction() {
//
//                public double[][] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
//                    double[][] J = new double[w * h][a.length];
//                    double d = 0;
//                    for (int i = 0; i < w; i++) {
//                        for (int j = 0; j < h; j++) {
//                            d = Math.exp(-a[1] * (x[i] - a[3]) * (x[i] - a[3])
//                                    - a[2] * (y[j] - a[4]) * (y[j] - a[4]));
//                            J[j * w + i][0] = d;
//                            d *= a[0];
//                            J[j * w + i][1] = -d * (x[i] - a[3]) * (x[i] - a[3]);
//                            J[j * w + i][2] = -d * (y[j] - a[4]) * (y[j] - a[4]);
//                            J[j * w + i][3] = d * 2 * a[1] * (x[i] - a[3]);
//                            J[j * w + i][4] = d * 2 * a[2] * (y[j] - a[4]);
//                            J[j * w + i][5] = 1;
//                        }
//                    }
//                    return J;
//                }
//            };
//        }
//
//        /*
//         * (non-Javadoc)
//         * @see org.apache.commons.math.analysis.MultivariateVectorialFunction#value(double[])
//         */
//        public double[] value(double[] a) throws FunctionEvaluationException, IllegalArgumentException {
//            double[] f = new double[w * h];
//            for (int i = 0; i < w; i++) {
//                for (int j = 0; j < h; j++) {
//                    f[j * w + i] = a[0] * Math.exp(-a[1] * (x[i] - a[3]) * (x[i] - a[3])
//                             - a[2] * (y[j] - a[4]) * (y[j] - a[4])) + a[5];
//                }
//            }
//            return f;
//        }
//    }

    /**
     * Fits a gauss with offset y = Ae^(-(x-C)^2/(2B^2)) + D to the given data.
     * 
     * @param x the independent variable data
     * @param y the dependent variable data
     * @param weights the weights of the measurements
     * @param startValues starting values for the parameters
     * @return array of calculated parameters ([A,B,C,D])
     * @throws OptimizationException
     * @throws FunctionEvaluationException
     * @throws IllegalArgumentException
     * @deprecated ues {@link FittingUtilities#lmLinear(double[], double[], double[], double[])}
     */
    @Deprecated
    public static double[] lmOffset(double[] x, double[] y, double[] weights, double[] startValues)
            throws OptimizationException, FunctionEvaluationException, IllegalArgumentException {
        LevenbergMarquardtOptimizer lmo = new LevenbergMarquardtOptimizer();
        //		GaussNewtonOptimizer lmo = new GaussNewtonOptimizer(true);
        lmo.setMaxIterations(500);
        GaussWithOffset f = new GaussWithOffset(x);
        VectorialPointValuePair pvp = lmo.optimize(f,y,weights,startValues);
        double[] a = pvp.getPointRef();
        a[1] = 1. / Math.sqrt(2 * a[1]);
        return a;
    }

    /**
     * Fits a gauss with linear function y = Ae^(-(x-C)^2/(2B^2)) + D + E x to the given data.
     * 
     * @param x the independent variable data
     * @param y the dependent variable data
     * @param weights the weights of the measurements
     * @param startValues starting values for the parameters
     * @return array of calculated parameters ([A,B,C,D,E])
     * @throws OptimizationException
     * @throws FunctionEvaluationException
     * @throws IllegalArgumentException
     */
    public static double[] lmLinear(double[] x, double[] y, double[] weights, double[] startValues)
            throws OptimizationException, FunctionEvaluationException, IllegalArgumentException {
        LevenbergMarquardtOptimizer lmo = new LevenbergMarquardtOptimizer();
        lmo.setInitialStepBoundFactor(0.3);
        lmo.setMaxIterations(1000);
        startValues[1] = 1 / (2 * startValues[1]);
        GaussWithLinear f = new GaussWithLinear(x);
        VectorialPointValuePair pvp = lmo.optimize(f,y,weights,startValues);
        double[] a = pvp.getPointRef();
        a[1] = 1. / Math.sqrt(2 * a[1]);
        return a;
    }

    /**
     * Fits a 2D gauss y = Ae^(-B(x-E)^2 - 2c(x-E)(y-F) - D(y-F)^2) + G to the given data.
     * 
     * @param x the independent variable data
     * @param y the dependent variable data
     * @param weights the weights of the measurements
     * @param startValues starting values for the parameters
     * @return array of calculated parameters ([A,B,C,D,E,F,G])
     * @throws OptimizationException
     * @throws FunctionEvaluationException
     * @throws IllegalArgumentException
     */
    public static double[] lmGauss2D(double[] x, double[] y, double[] z, double[] weights, double[] startValues)
            throws OptimizationException, FunctionEvaluationException, IllegalArgumentException {
        LevenbergMarquardtOptimizer lmo = new LevenbergMarquardtOptimizer();
        lmo.setInitialStepBoundFactor(0.1);
        lmo.setMaxIterations(1000);
        lmo.setCostRelativeTolerance(0.01);
        Gauss2D f = new Gauss2D(x,y);
        VectorialPointValuePair pvp = lmo.optimize(f,z,weights,startValues);
        System.out.println(lmo.getChiSquare());
        return pvp.getPointRef();
    }

    public static void main(String[] args) throws Exception {
        
        int N = 12;
        double[] x = new double[N];
        double[] y = new double[N];
        for (int i = 0; i < N; i++) {
            x[i] = i;
            y[i] = i;
        }
        Gauss2D g = new Gauss2D(x,y);
        
        double sx = 2;
        double sy = 1;
        double th = Math.PI/4;
        
        double aa = Math.cos(th) * Math.cos(th)/(2* sx * sx) + Math.sin(th) * Math.sin(th)/(2* sy * sy);
        double bb = - Math.sin(2 * th)/(4*sx*sx) + Math.sin(2*th)/(4*sy*sy);
        double cc = Math.cos(th) * Math.cos(th)/(2* sy * sy) + Math.sin(th) * Math.sin(th)/(2* sx * sx);
        
        System.out.println(aa);
        System.out.println(bb);
        System.out.println(cc);
        
        double[] a = {8, aa,bb,cc, 2,5, 1};
        double[] f = g.value(a);
        double[] w = new double[f.length];
        Arrays.fill(w,1);
        
        double[][] d = new double[N][N];
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {
                f[i*N+j] = f[i*N+j] + Math.random()  - 0.5;
                d[i][j] = f[i*N + j];
            }
            String str = Arrays.toString(d[i]);
            System.out.println(str.substring(1,str.length() - 1));
        }
                
        double[] dd = lmGauss2D(x,y,f,w,new double[] {4,2,1,1.1,1,1,0});
        System.out.println(Arrays.toString(dd));
        
        
//        //2,2,3,50,40
//        int w = 1000;
//        int h = w;
//        double[] x = new double[w];
//        double[] y = new double[h];
//        for (int i = 0; i < w; i++) {
//            x[i] = i;
//        }
//        for (int i = 0; i < w; i++) {
//            if (i == 200) continue;
//            double xx = 4 * (x[i] - 200);
//            y[i] = 2 + 100 * (1 + 2 * Math.cos(0.2 * (x[i] - 250))) * 
//                    BesselJ.valueAny(1,xx)*BesselJ.valueAny(1,xx)/(xx*xx);
//        }
//        double[] weights = new double[w];
//        Arrays.fill(weights,1);
//        double[] start = new double[]{1,90,4,200,2,0.2,250};
//        double[] data = lmInterferometer(false,x,y,weights,start,true);
//        System.out.println(Arrays.toString(data));
    }

    /**
     * Fits an interferometer function y = A + B sinc^2(C(x-D)) (1 + E cos(F(x-G))) to the given data.
     * 
     * @param x the independent variable data
     * @param y the dependent variable data
     * @param weights the weights of the measurements
     * @param startValues starting values for the parameters
     * @param useBesselFunction true to use bessel function, or false to use trigonometric
     * @return array of calculated parameters ([A,B,C,D,E,F,G])
     * @throws OptimizationException
     * @throws FunctionEvaluationException
     * @throws IllegalArgumentException
     */
    public static double[] lmInterferometer(boolean nelderMead, final double[] x, final double[] y, double[] weights,
            double[] startValues, boolean useBesselFunction) throws OptimizationException, FunctionEvaluationException, IllegalArgumentException {
        if (nelderMead) {
            MultiDirectional nm = new MultiDirectional();
            nm.setMaxIterations(1000);
            MultivariateRealFunction f = 
                    useBesselFunction ? new InterferometerCostFunctionBessel(x,y)
                            : new InterferometerCostFunctionTrigonometric(x,y);
            RealPointValuePair pair = nm.optimize(f,GoalType.MINIMIZE,startValues);
            return pair.getPointRef();
        } else {
            LevenbergMarquardtOptimizer lmo = new LevenbergMarquardtOptimizer();
            lmo.setMaxIterations(1000);
            DifferentiableMultivariateVectorialFunction f = 
                    useBesselFunction ? new InterferometerBessel(x)
                            : new InterferometerTrigonometric(x);
            VectorialPointValuePair pvp = lmo.optimize(f,y,weights,startValues);
            return pvp.getPointRef();
        }
    }

    /**
     * Fits a gauss with offset y = Ae^(-(x-C)^2/(2B^2)) to the given data.
     * 
     * @param x the independent variable data
     * @param y the dependent variable data
     * @param weights the weights of the measurements
     * @param startValues starting values for the parameters
     * @return array of calculated parameters ([A,B,C])
     * @throws OptimizationException
     * @throws FunctionEvaluationException
     * @throws IllegalArgumentException
     */
    @Deprecated
    public static double[] lmGauss(double[] x, double[] y, double[] weights, double[] startValues)
            throws OptimizationException, FunctionEvaluationException, IllegalArgumentException {
        LevenbergMarquardtOptimizer lmo = new LevenbergMarquardtOptimizer();
        lmo.setInitialStepBoundFactor(0.3);
        lmo.setMaxIterations(1000);
        startValues[1] = 1 / (2 * startValues[1]);
        Gauss f = new Gauss(x);
        VectorialPointValuePair pvp = lmo.optimize(f,y,weights,startValues);
        double[] a = pvp.getPointRef();
        a[1] = 1. / Math.sqrt(2 * a[1]);
        return a;
    }

    /**
     * Performs a linear fit of exponential function to the y data.
     * 
     * @param y
     * @param constY
     * @return
     */
    @Deprecated
    public static double[] fit(double[] y, double constY) throws MathException {
        double[][] A = new double[3][3];
        double[] b = new double[3];
        double val;
        double length = y.length;
        try {
            int i = 0;
            double v;
            double x;
            for (;;) {
                //normalize to eliminate overflow
                val = (y[i] - constY) / length;
                if (val <= 0) {
                    val = 0;
                } else {
                    val = Math.log(val);
                }
                //extra weight to boost large values instead of small
                v = y[i] - constY;
                //normalize to eliminate overflow
                x = i / length;
                A[0][0] += x * x * x * x * v;
                A[0][1] += x * x * x * v;
                A[0][2] += x * x * v;
                A[1][2] += x * v;
                A[2][2] += v;
                b[0] += val * x * x * v;
                b[1] += val * x * v;
                b[2] += val * v;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {}
        A[1][1] = A[0][2];
        A[1][0] = A[0][1];
        A[2][0] = A[0][2];
        A[2][1] = A[1][2];
        LUDecomposition lu = new LUDecomposition(A);
        double[] r = lu.solve(b);
        double sigma = Math.sqrt(-0.5 / r[0]) * length;
        double mean = -0.5 * r[1] * length / r[0];
        double amp = Math.exp(r[2] - 0.25 * r[1] * r[1] / r[0]) * length;
        return new double[]{amp,sigma,mean,constY,0};
    }

    /**
     * Checks if any of the values in the arrays is NaN.
     * 
     * @param values
     * @return
     */
    public static boolean isNaN(double[] values) {
        if (values == null) return true;
        for (int i = 0; i < values.length; i++) {
            if (Double.isNaN(values[i])) return true;
        }
        return false;
    }

    /**
     * Smooths the data with averaging.
     * 
     * @param x data to be smoothed
     * @return smoothed data
     */
    public static double[] smooth(double[] x) {
        return smooth(x,false);
    }

    /**
     * Smooths the values in the array with averaging and applying a low pass filter.
     * 
     * @param x the data to smooth
     * @param lowPassFilterOn true if low pass filtering should be used
     * @return smoothed array
     */
    public static double[] smooth(double[] x, boolean lowPassFilterOn) {
        int l = x.length;
        int n = 6;
        double[] xx = new double[l];
        for (int i = n; i < l; i++) {
            for (int j = 0; j < n; j++) {
                xx[i] += x[i - j];
            }
            xx[i] /= n;
        }
        for (int i = 0; i < n; i++) {
            xx[i] = xx[n];
        }
        if (!lowPassFilterOn) return xx;
        //get power of 2
        --l;
        l |= l >> 1;
        l |= l >> 2;
        l |= l >> 4;
        l |= l >> 8;
        l |= l >> 16;
        l++;
        double[] d = new double[l];
        double[] d1 = new double[l];
        System.arraycopy(xx,0,d1,0,xx.length);
        double[][] dd = new double[][]{d1,d};
        dd = FFT1D(dd,true);
        int q = (int)(dd[0].length * 0.05 + 0.5);
        int h = dd[0].length / 2;
        for (int i = q; i < h; i++) {
            dd[0][i] = 0;
            dd[0][h + i] = 0;
            dd[1][i] = 0;
            dd[1][h + i] = 0;
        }
        dd = FFT1D(dd,false);
        double[] a = new double[xx.length];
        System.arraycopy(dd[0],0,a,0,a.length);
        return a;
    }

    /**
     * @param x array of values. First array are real values, second imaginary
     * @param forward true for forward and false for inverse transform
     * @return Fourier transform of x
     * @throws Exception
     */
    public static double[][] FFT1D(double[][] x, boolean forward) {
        if (x == null) return null;
        if (x.length != 2 || x[0].length != x[1].length) {
            throw new IllegalArgumentException("bad x lengths");
        }
        int n = x[0].length;
        int n2 = 1;
        while (n2 < n) {
            n2 *= 2;
            if (n2 > n) {
                throw new IllegalArgumentException("x length must be power of 2");
            }
        }
        n2 = n / 2;
        double[][] temp = new double[2][n2];
        double angle = (-2.0 * Math.PI / n);
        if (!forward) angle = -angle;
        for (int i = 0; i < n2; i++) {
            temp[0][i] = Math.cos(i * angle);
            temp[1][i] = Math.sin(i * angle);
        }
        double[][] y = FFT1D(x,temp);
        if (!forward) {
            for (int i = 0; i < n; i++) {
                y[0][i] /= n;
                y[1][i] /= n;
            }
        }
        return y;
    }

    private static double[][] FFT1D(double[][] x, double[][] temp) {
        int n = x[0].length;
        int n2 = n / 2;
        int k = 0;
        int b;
        int c = 0;
        if (n == 1) {
            double[][] z1 = {{x[0][0]},{x[1][0]}};
            return z1;
        }
        b = (temp[0].length / n2);
        double[][] z = new double[2][n2];
        double[][] w = new double[2][n2];
        for (k = 0; k < n2; k++) {
            int k2 = 2 * k;
            z[0][k] = x[0][k2];
            z[1][k] = x[1][k2];
            w[0][k] = x[0][k2 + 1];
            w[1][k] = x[1][k2 + 1];
        }
        z = FFT1D(z,temp);
        w = FFT1D(w,temp);
        double[][] y = new double[2][n];
        for (k = 0; k < n2; k++) {
            y[0][k] = z[0][k];
            y[1][k] = z[1][k];
            double re = w[0][k] * temp[0][c] - w[1][k] * temp[1][c];
            double im = w[0][k] * temp[1][c] + w[1][k] * temp[0][c];
            w[0][k] = re;
            w[1][k] = im;
            y[0][k] += w[0][k];
            y[1][k] += w[1][k];
            y[0][k + n2] = z[0][k] - w[0][k];
            y[1][k + n2] = z[1][k] - w[1][k];
            c += b;
        }
        return y;
    }
}
